# -*- coding: utf-8 -*-
"""
This module provides access to the Scenario-Tool's REST API to create and run scenarios and manage models
"""
from pkg_resources import get_distribution, DistributionNotFound

from scenario_tool_interface.sti import ScenarioToolInterface, AccessLevel

from scenario_tool_interface.json2dynamind import Json2DynaMindXML
from scenario_tool_interface.dynamind2json import DynaMindXML2Json

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = 'scenario-tool-interface'
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = '0.0.2'
finally:
    del get_distribution, DistributionNotFound



import pytest
from scenario_tool_interface import ScenarioToolInterface
import json
import time
import os
# from .credentials import

__author__ = "Christian Urich"
__copyright__ = "Christian Urich"
__license__ = "mit"


USERNAME = os.environ['USERNAME']
PASSWORD = os.environ['PASSWORD']
# USERNAME_GUEST = os.environ['USERNAME_GUEST']
# PASSWORD_GUEST = os.environ['PASSWORD_GUEST']
# PASSWORD_ADMIN = os.environ['PASSWORD_ADMIN']
# USERNAME_ADMIN = os.environ['USERNAME_ADMIN']


def run_tutorial(queue="default"):
    sti = ScenarioToolInterface()
    # Login with your username and password
    sti.login(USERNAME, PASSWORD)

    project_id = sti.create_project()
    assert type(project_id) is int

    # Obtain region code
    region_id = sti.get_region( "melbourne")
    assert type(region_id) is int

    # Load geoson file
    with open("./resources/test.geojson", 'r') as file:
        geojson_file = json.loads(file.read())

    # Upload boundary
    geojson_id = sti.upload_geojson(geojson_file, project_id)

    # Set project parameters
    sti.update_project(project_id, {
        "name": "my project",
        "active": True,
        'region_id': region_id,
        "case_study_area_id": geojson_id,
    })

    # Add assessment models
    lst_model = sti.get_assessment_model( "Land Surface Temperature")
    assert type(lst_model) is int

    # Set assessment models
    sti.set_project_assessment_models(project_id, [{"assessment_model_id": lst_model}])

    # Create and run baseline
    baseline_id = sti.create_scenario(project_id, None)
    assert type(baseline_id) is int
    sti.execute_scenario(baseline_id)

    # Scenarios are executed asynchronous.
    while True:
        # A status code smaller than 7 means the simulation is still executed.
        r = sti.check_status(baseline_id)
        status = r["status"]
        if status > 6:
            break
        time.sleep(1)


    # Print a list of available nodes
    sti.show_nodes()

    #Get residential node id
    nodes = sti.get_nodes()
    n_id = -1
    for n in nodes:
        if n["name"] == "Residential":
            n_id = n["id"]


    # Nodes are defined as below
    residential_node = {
        "node_type_id": n_id,
        "area": geojson_id,
        "parameters":
            {
                "dance4water_building.site_coverage": 0.6,
                "dance4water_number_of_trees.equation": 1,
                "dance4water_tree_spacing.equation": 22
            }
    }

    nodes = []
    # Several nodes can are combined to a workflow be adding them to a vector. The
    # nodes are executed in the order the are added
    nodes.append(residential_node)

    # Scenarios need a parent. In this case we use the base line scenario created before
    baseline_scenario_id = sti.get_baseline(project_id)

    # Crate a new scenario
    scenario_id = sti.create_scenario( project_id, baseline_scenario_id, "my new scenario")
    assert type(scenario_id) is int
    # Set workflow
    sti.set_scenario_workflow(scenario_id, nodes)

    # Execute scenario
    sti.execute_scenario(scenario_id, queue=queue)

    # Scenarios are executed asynchronous
    while True:
        # A status code smaller than 7 means the simulation is still executed.
        r = sti.check_status(scenario_id)
        status = r["status"]
        if status > 6:
            break

        time.sleep(1)

    while True:
        r = sti.run_query(scenario_id,
                          "SELECT avg(tree_cover_fraction) as tf from micro_climate_grid")

        if r['status'] == 'loaded':
            # Break the loop when query is loaded
            break
        time.sleep(1)
    print(r['data'])
    assert r['data'][0]['tf'] == 0.0696807189475445


    # Should through an exception because access
    with pytest.raises(Exception) as e_info:
        sti.get_project_databases(project_id)


# def test_login():
#     assert type(sti.login(USERNAME, PASSWORD)) is str
#     with pytest.raises(Exception) as e_info:
#         sti.login(USERNAME, "passwod")
#
# def test_guest_login():
#     assert type(sti.login(USERNAME_GUEST, PASSWORD_GUEST)) is str
#
# def test_admin_login():
#     assert type(sti.login(USERNAME_ADMIN, PASSWORD_ADMIN)) is str
#
# def test_get_region():
#     token = sti.login(USERNAME, PASSWORD)
#
#     assert type(sti.get_region(token, "melbourne")) is int
#
#     with pytest.raises(Exception) as e_info:
#         sti.login(sti.get_region(token, "melbourn"))
#
#
# def test_get_assessment_model():
#     token = sti.login(USERNAME, PASSWORD)
#
#     assert type(sti.get_assessment_model(token, "Land Surface Temperature")) is int
#
#     with pytest.raises(Exception) as e_info:
#         sti.login(sti.get_assessment_model(token, "Land Surface Temperatur"))
#
#
# # def test_nodes():
# #     # Login with your username and password
# #     token = sti.login(USERNAME, PASSWORD)
# #
# #     assert type(token) is str
# #
# #     model_id = sti.upload_dynamind_model(token, "test node", "resources/nodes/test_node.dyn")
# #
# #     assert type(model_id) is int
# #
# #     node_id = sti.create_node(token, "resources/nodes/test_node.json", model_id)
# #
# #     assert type(node_id) is int
# #
# #     n_id = sti.get_node_id(token, "test node")
# #
# #     assert type(n_id)
# #
# #     with pytest.raises(Exception) as e_info:
# #         sti.get_node_id(token, "test nod")
# #
# #     model_id = sti.upload_dynamind_model(token, "test node", "resources/nodes/test_node.dyn")
# #
# #     assert type(model_id) is int
# #
# #     node_version_id = sti.update_node(token, node_id, "resources/nodes/test_node.json", model_id)
# #
# #     assert type(node_version_id) is int
# #
# #     sti.deactivate_node(token, node_id)
# #
# #     with pytest.raises(Exception) as e_info:
# #         sti.get_node_id(token, "test node")
#
#
#
# def test_assessment_models():
#     # Login with your username and password
#     token = sti.login(USERNAME, PASSWORD)
#
#     assert type(token) is str
#
#     model_id = sti.upload_dynamind_model(token, "test node", "resources/assessment_nodes/test_assessment_model.dyn")
#
#     assert type(model_id) is int
#
#     node_id = sti.create_assessment_model(token, "resources/assessment_nodes/test_assessment_model.json", model_id)
#
#     assert type(node_id) is int
#
#     model_id = sti.upload_dynamind_model(token, "test node", "resources/assessment_nodes/test_assessment_model.dyn")
#
#     assert type(model_id) is int
#
#     node_version_id = sti.update_assessment_model(token, node_id, "resources/assessment_nodes/test_assessment_model.json", model_id)
#
#     assert type(node_version_id) is int
#
#
def test_run_tutorial():
    run_tutorial()
#
# # def test_run_tutorial_slow():
# #     run_tutorial(queue="slow")
#
# def test_download():
#     # Login with your username and password
#     token = sti.login(USERNAME_ADMIN, PASSWORD_ADMIN)
#
#     assert type(token) is str
#     # Create a new project
#     project_id = sti.create_project(token)
#     assert type(project_id) is int
#
#     # Obtain region code
#     region_id = sti.get_region(token, "melbourne")
#     assert type(region_id) is int
#
#     # Load geoson file
#     with open("resources/test_small.geojson", 'r') as file:
#         geojson_file = json.loads(file.read())
#
#     # Upload boundary
#     geojson_id = sti.upload_geojson(token, geojson_file, project_id)
#
#     # Set project parameters
#     sti.update_project(token, project_id, {
#         "name": "my project small",
#         "active": True,
#         'region_id': region_id,
#         "case_study_area_id": geojson_id,
#     })
#
#     # Add assessment models
#     lst_model = sti.get_assessment_model(token, "Land Surface Temperature")
#     assert type(lst_model) is int
#
#     # Set assessment models
#     sti.set_project_assessment_models(token, project_id, [{"assessment_model_id": lst_model}])
#
#     # Create and run baseline
#     baseline_id = sti.create_scenario(token, project_id, None)
#     assert type(baseline_id) is int
#     sti.execute_scenario(token, baseline_id)
#
#     # Scenarios are executed asynchronous.
#     while True:
#         # A status code smaller than 7 means the simulation is still executed.
#         r = sti.check_status(token, baseline_id)
#         status = r["status"]
#         if status > 6:
#             break
#         time.sleep(1)
#
#     # Print a list of available nodes
#     sti.show_nodes(token)
#
#     # Get residential node id
#     nodes = sti.get_nodes(token)
#     n_id = -1
#     for n in nodes:
#         if n["name"] == "Residential":
#             n_id = n["id"]
#
#     # Nodes are defined as below
#     residential_node = {
#         "node_type_id": n_id,
#         "area": geojson_id,
#         "parameters":
#             {
#                 "dance4water_building.site_coverage": 0.6,
#                 "dance4water_number_of_trees.equation": 1,
#                 "dance4water_tree_spacing.equation": 22
#             }
#     }
#
#     nodes = []
#     # Several nodes can are combined to a workflow be adding them to a vector. The
#     # nodes are executed in the order the are added
#     nodes.append(residential_node)
#
#     # Scenarios need a parent. In this case we use the base line scenario created before
#     baseline_scenario_id = sti.get_baseline(token, project_id)
#
#     # Crate a new scenario
#     scenario_id = sti.create_scenario(token, project_id, baseline_scenario_id, "my new scenario")
#     assert type(scenario_id) is int
#     # Set workflow
#     sti.set_scenario_workflow(token, scenario_id, nodes)
#
#     # Execute scenario
#     sti.execute_scenario(token, scenario_id)
#
#     # Scenarios are executed asynchronous
#     while True:
#         # A status code smaller than 7 means the simulation is still executed.
#         r = sti.check_status(token, scenario_id)
#         status = r["status"]
#         if status > 6:
#             break
#
#         time.sleep(1)
#
#     sti.get_project_databases(token, project_id, "/tmp")
#     assert os.path.exists(f"/tmp/{project_id}.zip")
#
# def test_node_access_level():
#     # Login with your username and password
#     token = sti.login(USERNAME, PASSWORD)
#     assert type(token) is str
#
#     #Upload test model
#     model_id = sti.upload_dynamind_model(token, "test node", "resources/nodes/test_node.dyn")
#
#     assert type(model_id) is int
#
#     n_id = sti.create_node(token, "resources/nodes/test_node.json", model_id)
#     assert type(n_id) is int
#
#     token = sti.login(USERNAME_GUEST, PASSWORD_GUEST)
#     assert type(token) is str
#
#     # Create a new project
#     project_id = sti.create_project(token)
#     assert type(project_id) is int
#
#     # Obtain region code
#     region_id = sti.get_region(token, "melbourne")
#     assert type(region_id) is int
#
#     # Load geoson file
#     with open("resources/test_small.geojson", 'r') as file:
#         geojson_file = json.loads(file.read())
#
#     # Upload boundary
#     geojson_id = sti.upload_geojson(token, geojson_file, project_id)
#
#     # Set project parameters
#     sti.update_project(token, project_id, {
#         "name": "my project small",
#         "active": True,
#         'region_id': region_id,
#         "case_study_area_id": geojson_id,
#     })
#
#     # Add assessment models
#     lst_model = sti.get_assessment_model(token, "Land Surface Temperature")
#     assert type(lst_model) is int
#
#     # Set assessment models
#     sti.set_project_assessment_models(token, project_id, [{"assessment_model_id": lst_model}])
#
#     # Create and run baseline
#     baseline_id = sti.create_scenario(token, project_id, None)
#     assert type(baseline_id) is int
#     sti.execute_scenario(token, baseline_id)
#
#     # Scenarios are executed asynchronous.
#     while True:
#         # A status code smaller than 7 means the simulation is still executed.
#         r = sti.check_status(token, baseline_id)
#         status = r["status"]
#         if status > 6:
#             break
#         time.sleep(1)
#
#
#
#     # Nodes are defined as below
#     test_nodes = {
#         "node_type_id": n_id,
#         "area": geojson_id,
#         "parameters":
#             {
#             }
#     }
#     nodes = []
#     # Several nodes can are combined to a workflow be adding them to a vector. The
#     # nodes are executed in the order the are added
#     nodes.append(test_nodes)
#
#     # Scenarios need a parent. In this case we use the base line scenario created before
#     baseline_scenario_id = sti.get_baseline(token, project_id)
#
#     # Crate a new scenario
#     scenario_id = sti.create_scenario(token, project_id, baseline_scenario_id, "my new scenario")
#     assert type(scenario_id) is int
#     # Set workflow
#     with pytest.raises(Exception) as e_info:
#         sti.set_scenario_workflow(token, scenario_id, nodes)
#
#     # Give me access to the node
#     sti.set_node_access_level(sti.login(USERNAME, PASSWORD), n_id, sti.AccessLevel.DEMO.value)
#     sti.set_scenario_workflow(token, scenario_id, nodes)
#
#     # Execute scenario
#     sti.execute_scenario(token, scenario_id)
#
#     # Remove node
#     sti.deactivate_node(token,  n_id)
#
#     with pytest.raises(Exception) as e_info:
#         sti.get_node_id(token, "test node")
#
#     # Scenarios are executed asynchronous
#     while True:
#         # A status code smaller than 7 means the simulation is still executed.
#         r = sti.check_status(token, scenario_id)
#         status = r["status"]
#         if status > 6:
#             break
#
#         time.sleep(1)








#!/usr/bin/env python
# -*- coding: utf-8 -*-

import requests
import json
import time
import enum


class AccessLevel(enum.Enum):
    DEMO = 1
    PARTICIPANT = 2
    CONSULTANT = 3
    ADMIN = 4
    SUPERADMIN = 5


class ScenarioToolInterface:

    def __init__(self, api_url="https://stable-api.dance4water.org/api",
                 results_url="https://stable-sql.dance4water.org/resultsdb/"):
        self.api_url = api_url
        self.results_url = results_url

        self.authenticated = False
        self.token = None
        self._cache = {}

    def login(self, username: str, password: str):
        """
        :param username: registered user name (email address)
        :param password: top secret password
        """
        counter = 0
        while True:
            r = requests.post(self.api_url + "/user/login/", json={'username': username,
                                                                   'password': password})
            counter += 1
            if r.status_code == 200:
                self.token = r.json()["access_token"]
                self.authenticated = True
                return
            if counter > 4:
                raise Exception(f"Unable to login status {r.status_code}")

            else:
                time.sleep(2)

    def _get(self, url):

        if not self.authenticated or self.token is None:
            raise Exception(f"User not authenticated, GET FAILED, {url}")

        headers = {"Authorization": "Bearer " + self.token}
        return requests.get(url, headers=headers)

    def _put(self, url, data={}):

        if not self.authenticated or self.token is None:
            raise Exception(f"User not authenticated, PUT FAILED, {url}")

        headers = {"Authorization": "Bearer " + self.token}
        return requests.put(url, json=data, headers=headers)

    def _post(self, url, data={}):

        if not self.authenticated or self.token is None:
            raise Exception(f"User not authenticated, {url}")

        headers = {"Authorization": "Bearer " + self.token}
        return requests.post(url, json=data, headers=headers)

    def db_name(self, simulation_id):
        return simulation_id

    def run_query(self, scenario_id, query):
        """
        Run a query on the scenario spatialite database. The database supports SQLite and Spatialite commands
        Only read access is supported! To find out which data are stored in the database read the dynamind_table_definitions

        :param scenario_id: scenario id
        :param query: SQL query
        :type scenario_id: int
        :type query str
        :return: query result
        :rtype: dict
        """

        simulation_id = self.get_database_id(scenario_id)

        data = {'db_name': self.db_name(simulation_id),
                'query': query}
        r = requests.post(self.results_url, data=data)
        if r.status_code == 200:
            result = r.json()
            return result
        raise Exception(f"Unable to run query {r.status_code}")

    def add_node(self, node_data):
        return self._post(self.api_url + "/sm_node", node_data)

    def update_sm_node(self, node_id, node_data):
        return self._post(self.api_url + "/sm_node/" + str(node_id) + "/versions", node_data)

    def add_model(self, name, src):
        return self._post(self.api_url + "/models", {"name": name, "model_src": src})

    def get_models(self):
        return self._get(self.api_url + "/models/")

    def get_database_id(self, scenario_id):
        """
        Returns database ID to run data analysis

        :param scenario_1_id: scenario id
        :return: data base id needed for query
        """
        cache_key = f"get_database_id{scenario_id}"
        if cache_key in self._cache:
            return self._cache[cache_key]

        r = self.get_simulations(scenario_id)
        if r.status_code != 200:
            raise Exception(f"Unable to obtain scenarios {r.status_code}, {r.json()}")

        sims = r.json()

        for s in sims["simulations"]:
            sim = json.loads(s)
            if sim["simulation_type"] == "PERFORMANCE_ASSESSMENT":
                self._cache[cache_key] = sim["id"]
                return self._cache[cache_key]


    def create_project(self):
        """
        Creates a new project

        :return: project id
        :rtype: int
        """

        r = self._post(self.api_url + "/projects")

        if r.status_code == 200:
            return r.json()["id"]

        raise Exception(f"Creation of project failed {r.status_code}")

    def get_project(self, project):
        return self._get(self.api_url + "/projects/" + str(project))

    def get_projects(self):
        return self._get(self.api_url + "/projects/")

    def update_project(self, project, data):
        r = self._put(self.api_url + "/projects/" + str(project), data)

        if r.status_code == 200:
            return

        raise Exception(f"Updating project failed {r.status_code}, {r.json()}")

    def get_assessment_models(self):
        r = self._get(self.api_url + "/assessment_models/")

        if r.status_code == 200:
            return r.json()
        raise Exception(f"Failed to get performance assessment model {r.status_code}, {r.json()}")

    def get_assessment_model(self, model_name, owner_id=None):
        """
        Returns assessment model id

        :param model_name: Model Name. Currently supported are Land Surface Temperature and Target
        :param owner_id: If owner_id is set only models owned by the user will be returned
        :type model_name: str
        :type owner_id: int
        :return: model_id
        :rtype: int
        """
        r = self.get_assessment_models()

        models = r["assessment_models"]
        filtered_nodes = []
        for model in models:
            if model['name'] == model_name:
                if owner_id and model['creator'] == owner_id:
                    filtered_nodes.append(model)
                elif not owner_id:
                    filtered_nodes.append(model)
        if len(filtered_nodes) == 0:
            raise Exception(f"Performance assessment model {model_name} not found")

        if len(filtered_nodes) == 1:
            return filtered_nodes[0]["id"]

        # if multiple performance assessment models have been found return the one the user owns
        for n in filtered_nodes:
            if n["creator"] == self.get_my_status()["user_id"]:
                return n["id"]

        return filtered_nodes[0]["id"]

    def set_project_assessment_models(self, project, models):
        r = self._put(self.api_url + "/projects/" + str(project) + "/models", models)
        if not r.status_code == 200:
            raise Exception(f"Unable to set assessment model {r.status_code}, {r.json()}")

    def set_project_data_model(self, project, model):
        r = self._post(self.api_url + "/projects/" + str(project) + "/data_model", model)
        if not r.status_code == 200:
            raise Exception(f"Unable to set project data model {r.status_code}, {r.json()}")


    def share_project(self, project, username):
        """
        Share a project

        :param project: project id
        :param username: share user name
        """

        data = {"username": username}

        r = self._post(self.api_url + "/projects/" + str(project) + "/share", data)

        if r.status_code == 200:
            return

        raise Exception(f"Unable to share project {r.status_code}, {r.json()}")

    def archive_project(self, project):
        """
        Share a project

        :param project: project id
        :param username: share user name
        """

        data = {}

        r = self._post(self.api_url + "/projects/" + str(project) + "/archive", data)

        if r.status_code == 200:
            return

        raise Exception(f"Unable to archive project {r.status_code}")


    def create_scenario(self, project, parent, name="Baseline"):
        """
        Creates a new scenario. The provides the shell for the new scenarios. Scenario are derived from the base line
        or any other scenario in the project. To modify the environment workflow may be defined and executed.

        :param project: project id
        :param parent: parent scenario id
        :param name: name of scenario
        :type int
        :type int
        :type str
        :return: scenario id
        :rtype: int
        """

        data = {"project_id": project, "name": name}
        if parent is not None:
            data["parent"] = parent

        r = self._post(self.api_url  + "/scenario/", data)

        if r.status_code == 200:
            return r.json()["id"]

        raise Exception(f"Unable to create scenario {r.status_code}, {r.json()}")

    def set_scenario_workflow(self, scenario_id, node_data):
        """
        Set the workflow for a scenario. The workflow is defined by a series of nodes defined by the node_data
        The node_data have following structure

        .. code-block::

            [{
               name: node_id,
               area: geojson_id,
               parameters: {

                parameter.value = 1,
                paramter2.value = 2,

               }
            },
            ...
            ]

        The nodes in the workflow are executed as defined in the data structure

        :param scenario_id: scenario id
        :param node_data: node data see above
        :type scenario_id: int
        :type node_data: list

        """
        r = self._post(self.api_url + "/scenario/" + str(scenario_id) + "/nodes", node_data)
        if r.status_code == 200:
            return

        raise Exception(f"Something went wrong when adding the nodes {r.status_code} {r.json()}")

    def get_scenario_workflow_nodes(self, scenario_id=None):
        if scenario_id is None:
            return self._get(self.api_url + "/sm_node/")
        return self._get(self.api_url + "/scenario/" + str(scenario_id) + "/nodes")

    def create_dash_tile_template(self, json):
        """
        Define tile template (see doc)

        :param json:
        :return: dash_tile_template_id
        :rtype: int
        """

        r = self._post(self.api_url + "/dash_tile_template/", {"json": json})

        if r.status_code == 200:
            return r.json()["id"]

        raise Exception(f"Unable to create template {r.status_code}")

    def edit_dash_tile_template(self, template_id, json=None, active=True):
        """
        Define tile template (see doc)

        :param json:
        :return: dash_tile_template_id
        :rtype: int
        """
        data = {}
        if json:
            data["json"] = json
        data["active"] = active

        r = self._post(self.api_url + "/dash_tile_template/" + str(template_id), data)

        if r.status_code == 200:
            return r.json()["id"]

        raise Exception(f"Unable to create template {r.status_code}")

    def create_query_template(self, json, access_level):
        """
        Define tile template (see doc)

        :param json:
        :param access_level
        :type geojson: str
        :type access_level: int
        :return: query_template_id
        :rtype: int
        """

        r = self._post(self.api_url + "/query_prototypes/", {"json": json, "access_level": access_level})

        if r.status_code == 200:
            return r.json()["id"]

        raise Exception(f"Unable to create template {r.status_code}")

    def link_dash_tile_assessment_model(self, dash_tile_template_id, assessment_model_id):
        """
        Link dash tile with assessment model

        :param dash_tile_template_id:
        :param assessment_model_id:
        :type dash_tile_template_id: int
        :type assessment_model_id: int
        :return: None
        """

        r = self._post(self.api_url + "/dash_tile_template_assessment_model/",
                       {"dash_tile_template_id": dash_tile_template_id,
                        "assessment_model_id": assessment_model_id})
        if r.status_code == 200:
            return

        raise Exception(f"Unable to link template {r.status_code}")

    def link_query_prototype_assessment_model(self, query_prototype_id, assessment_model_id):
        """
        Link query prototype with assessment model

        :param query_prototype_id:
        :param assessment_model_id:
        :type query_prototype_id: int
        :type assessment_model_id: int
        :return: None
        """

        r = self._post(self.api_url + "/query_template_assessment_model/",
                       {"query_prototype_id": query_prototype_id,
                        "assessment_model_id": assessment_model_id})
        if r.status_code == 200:
            return

        raise Exception(f"Unable to link template {r.status_code}")

    def link_dash_tile_query_prototype(self, dash_tile_template_id , query_prototype_id):
        """
        Link query prototype with assessment model

        :param dash_tile_template_id:
        :param query_prototype_id:
        :type dash_tile_template_id: int
        :type query_prototype_id: int
        :return: None
        """

        r = self._post(self.api_url + "/dash_tile_template_query_prototype/",
                       {"dash_tile_template_id": dash_tile_template_id,
                        "query_prototype_id": query_prototype_id})
        if r.status_code == 200:
            return

        raise Exception(f"Unable to link template {r.status_code}")

    def upload_geojson(self, geojson, project_id, name="casestudyarea"):
        """
        Upload a geojson file and return id

        :param geojson: geojson file
        :param project_id: project the node will be assigned to
        :param name: added option to set name of geojson file default is set to casestudyarea
        :type geojson: str
        :type name: str
        :type project_id: int
        :return: geojson id
        :rtype: int
        """

        if 'name' in geojson:
            del geojson['name']
        r = self._post(self.api_url + "/geojson/", {"project_id": project_id, "geojson": geojson, "name": name})

        if r.status_code == 200:
            return r.json()["id"]

        raise Exception(f"Unable to upload file {r.status_code}, {r.json()}")

    def create_datasource(self,  project_id, name, source_type="file", upload_id=None):
        """
        Upload a geojson file and return id

        :param geojson: geojson file
        :param project_id: project the node will be assigned to
        :param name: added option to set name of geojson file default is set to casestudyarea
        :type geojson: str
        :type name: str
        :type project_id: int
        :return: geojson id
        :rtype: int
        """

        r = self._post(self.api_url + "/data_source/",
                       {"project_id": project_id,
                        "source_type": source_type,
                        "upload_id": upload_id,
                        "name": name})

        if r.status_code == 200:
            return r.json()["id"]

        raise Exception(f"Unable to create datasource {r.status_code}, {r.json()}")

    def upload_file(self, project_id, file):
        """
        Upload a  file and return id

        :param file: path to file
        :param project_id: project the file will be assigned to
        :type file: str
        :type name: str
        :type project_id: int
        :rtype: int
        """

        if not self.authenticated or self.token is None:
            raise Exception(f"User not authenticated, {url}")

        headers = {"Authorization": "Bearer " + self.token}

        files = {'file': open(file, 'rb')}
        r = requests.post(self.api_url + f"/upload/{project_id}", headers=headers, files=files)

        if r.status_code == 200:
            return r.json()["id"]

        raise Exception(f"Unable to upload file {r.status_code}, {r.json()}")


    def get_region(self, region_name: str) -> int:
        """
        Returns region currently supported is Melbourne

        :param region_name: region id
        :return: region id
        """

        r = self._get(self.api_url + "/regions/")
        if not r.status_code == 200:
            raise Exception(f"Unable to get region {r.status_code}, {r.json()}")
        regions = r.json()
        melbourne_region_id = None
        for region in regions:
            if region["name"].lower() == region_name:
                melbourne_region_id = region["id"]
                break

        if melbourne_region_id is None:
            raise Exception(f"Could not find ' {region_name}")

        return melbourne_region_id

    def get_regions(self) -> requests.Response:
        return self._get(self.api_url + "/regions/")

    def execute_scenario(self, scenario: int, queue="default") -> requests.Response:
        """

        :param scenario: id of scenario to be executed
        :param queue: optional parameter to define queue
        :rtype: int
        :rtype: str
        """
        r = self._post(f'{self.api_url}/scenario/{scenario}/execute?queue={queue}')
        if r.status_code != 200:
            raise Exception(f"Unable to execute scenario {r.status_code}, {r.json()}")

    def get_geojsons(self, project: int):
        return self._get(self.api_url + "/geojson/" + str(project))

    def check_status(self, scenario: int):
        """
        Return status of current simulation.

        returns:

        .. code-block::

            {
               status: status code (int),
               status_text: status description
            }

            // CREATED = 1
            // BASE_IN_QUEUE = 2
            // BASE_RUNNING = 3
            // BASE_COMPLETE = 4
            // PA_IN_QUEUE = 5
            // PA_RUNNING = 6
            // PA_COMPLETE = 7
            // COMPLETE = 8

        :param scenario: scenario id
        :type scenario: int
        :return: scenario status
        :rtype: dict

        """
        r = self._get(self.api_url + "/scenario/" + str(scenario) + "/status")
        if r.status_code != 200:
            raise Exception(f"Unable to get status {r.status_code}, {r.json()}")
        return r.json()

    def get_scenario(self, scenario: int):
        return self._get(self.api_url + "/scenario/" + str(scenario))

    def get_simulations(self, scenario: int):
        return self._get(self.api_url + "/scenario/" + str(scenario) + "/simulations")

    def upload_dynamind_model(self, name: str, filename: str) -> int:
        """
        Uploads a new model to the server

        :param name: model name
        :param filename: dynamind file
        :type str
        :return: model_id
        :rtype: int
        """
        with open(filename, 'r') as file:
            data = file.read().replace('\n', '')

        r = self.add_model(name, data)

        model_id = r.json()["model_id"]

        return model_id

    def show_node_versions(self, node_id: int):
        return self._get(self.api_url + "/sm_node/" + str(node_id))

    def create_node(self, filename, model_id=None, access_level=AccessLevel.SUPERADMIN.value):
        """
        Create a new node

        :param filename: point to json file containing the node description
        :param model_id: model id in json file will be replaced by this. If not set model_id from json file
        :param access_level: access level of node
        :type str
        :type int
        :type int
        :return: node_id
        :rtype: int
        """
        with open(filename) as json_file:
            node_data = json.load(json_file)

        if model_id is not None:
            node_data["models"][0]["id"] = model_id
            node_data["access_level"] = access_level
        r = self.add_node(node_data)

        if r.status_code == 200:
            result = r.json()
            return result["node_id"]
        raise Exception(f"Unable to add node {r.status_code}, {r.json()}")

    def update_node(self, node_id, filename, model_id=None, access_level=AccessLevel.SUPERADMIN.value):
        """
        Update an existing node

        :param node_id: id of node
        :param filename: point to json file containing the node description
        :param model_id: model id in json file will be replaced by this. If not set model_id from json file
        :param access_level: access level of node
        :type str
        :type str
        :type int
        :type int
        :return: node_id
        :rtype: int
        """
        with open(filename) as json_file:
            node_data = json.load(json_file)

        if model_id is not None:
            node_data["models"][0]["id"] = model_id
            node_data["access_level"] = access_level
        r = self.update_sm_node(node_id, node_data)

        if r.status_code == 200:
            result = r.json()
            return result["node_version_id"]
        raise Exception(f"Unable to update node {r.status_code}, {r.json()}")

    def set_node_access_level(self, node_id, access_level):
        """
        Set the access level of the parent node

        :param node_id: node id
        :param access_level: access level (see enum)
        :type node_id: int
        :type access_level: int
        """
        r = self._post(f"{self.api_url}/sm_node/{node_id}", {"access_level": access_level})
        if r.status_code == 200:
            return
        raise Exception(f"Could not update access level node {r.status_code}, {r.json()}")

    def set_node_properties(self, node_id, properties):
        """
        Set the access level of the parent node

        :param node_id: node id
        :param properties: dict
        :type node_id: int
        :type properties: dict

          .. code-block::

            {
               "active": true or false,
               "access_level": see enum,
               "tags": '["urban from", ...]' // Array as string with tags
               "description": "node description"
               "version_tag": "0.0.1" // Version Tag
            }

        returns:

        """
        r = self._post(f"{self.api_url}/sm_node/{node_id}", properties)
        if r.status_code == 200:
            return
        raise Exception(f"Could not update node properties node {r.status_code}, {r.json()}")

    def set_model_access_level(self, node_id, access_level):
        """
        Set the access level of the performance assessment model

        :param node_id: model id
        :param access_level: access level (see enum)
        :type node_id: int
        :type access_level: int
        """
        r = self._post(f"{self.api_url}/assessment_models/{node_id}", {"access_level": access_level})
        if r.status_code == 200:
            return
        raise Exception(f"Could not update access level performance assessment model {r.status_code}")

    def deactivate_node(self, node_id):
        """
        Deactivate node

        :param node_id: node id
        :type node_id: int
        """
        r = self._post(f"{self.api_url}/sm_node/{node_id}", {"active": False})
        if r.status_code == 200:
            return
        raise Exception(f"Could not deactivate node {r.status_code}, {r.json()}")

    def deactivate_assessment_model(self, node_id):
        """
        Deactivate assessment model

        :param node_id: node id
        :type node_id: int
        """
        r = self._post(f"{self.api_url}/assessment_models/{node_id}", {"active": False})
        if r.status_code == 200:
            return
        raise Exception(f"Could not deactivate assessment model {r.status_code}, {r.json()}")

    def get_baseline(self, project_id):
        """
        Get a projects baseline scenario id

        :param project_id: Project ID
        :type project_id: int
        :return: baseline scenario id
        :rtype: int
        """
        r = self.get_project(project_id)
        scenarios = r.json()["scenarios"]
        scene = next(item for item in scenarios if item["parent"] is None)
        baseline_id = scene["id"]
        return baseline_id

    def get_city_boundray(self, project_id):
        """
        Return a cities geojson boundary id

        :param project_id: project ID
        :type project_id: int
        :return: geojson boundary id
        :rtype: int
        """

        r = self.get_geojsons(project_id)
        geojsons = r.json()
        geojson_city_id = geojsons["geojsons"][0]["id"]
        return geojson_city_id

    def show_nodes(self):
        """
        Prints list of available nodes

        """
        r = self.get_scenario_workflow_nodes()
        if not r.status_code == 200:
            raise Exception(f"Could not get scenario workflow nodes")

        smnodes = r.json()["scenario_maker_nodes"]
        for node in smnodes:
            print(node["id"], node["name"])

    def get_nodes(self):
        """
        Return list of available nodes

        :return: returns a dict of all scenario
        :rtype: dict
        """
        r = self.get_scenario_workflow_nodes()
        if not r.status_code == 200:
            raise Exception(f"Could not get scenario workflow nodes")

        smnodes = r.json()["scenario_maker_nodes"]

        return smnodes

    def show_scenarios(self, project_id):
        """
        Prints a list of the scenarios in a project

        :param project_id: project id
        :type project_id: int
        """
        r = self.get_project(project_id)
        if not r.status_code == 200:
            raise Exception(f"Could not get scenario workflow nodes {r.status_code}, {r.json()}")

        scenarios = r.json()["scenarios"]
        for s in scenarios:
            print(s["id"], s["status"], s["name"])

    def get_scenarios(self, project_id):
        """
        Get a list of scenarios in a project

        :param project_id: project ID
        :type int
        :return: returns a dict of all scenario
        :rtype: dict
        """
        r = self.get_project(project_id)
        if not r.status_code == 200:
            raise Exception(f"Could not get scenario workflow nodes {r.status_code}, {r.json()}")

        scenarios = r.json()["scenarios"]
        return scenarios

    def download_geojson(self, scenario_id, layer_name):
        """
        Download layer as geojson file

        :param scenario_id: scenario id
        :param layern_name:
        :type int
        :type string
        :return: geojson string
        """

        r = self._get(self.api_url + "/scenario/" + str(scenario_id) + "/layer/" + str(layer_name))
        if not r.status_code == 200:
            raise Exception(f"Could not download result {r.status_code}, {r.json()}")

        return r.text

    def show_log(self, scenario_id):
        """
        Print scenario log file

        :param scenario_id: scenario id
        :type scenario_id: int
        """
        r = self.get_simulations( scenario_id)

        if not r.status_code == 200:
            raise Exception(f"Could not get scenario log {r.status_code}, {r.json()}")

        sims = r.json()

        for s in sims["simulations"]:
            database_id = json.loads(s)["id"]

        for s in sims["simulation_instances"]:
            print(json.loads(s)["id"], json.loads(s)["progress"], json.loads(s)["heartbeat"], json.loads(s)["log"])

        return database_id

    def get_node_id(self, name, owner_id=None):
        """
        Return node id to be used in simulation. If multiple nodes with the same id are identified the first node
        belonging to the user is returned first. If owner_id is set only the node owned by the user is returned

        :param name: node name
        :param owner_id: user_id
        :type name: str
        :type owner_id: int
        :return: node_id
        :rtype int
        """
        nodes = self.get_nodes()
        filtered_nodes = []
        for n in nodes:
            if n['name'] == name:
                if owner_id and n['creator'] == owner_id:
                    filtered_nodes.append(n)
                elif not owner_id:
                    filtered_nodes.append(n)
        if len(filtered_nodes) == 0:
            raise Exception(f"Node  {name} not found")

        if len(filtered_nodes) == 1:
            return filtered_nodes[0]["id"]
        # if multiple nodes return the one the user owns
        for n in filtered_nodes:
            if n["creator"] == self.get_my_status()["user_id"]:
                return n["id"]

        return filtered_nodes[0]["id"]

    def create_assessment_model(self, filename, model_id=None):
        """
        Creates a new assessment model and a default version tagged as 0.0.1
        the data must be of the shape:

        :param filename: filename of json file (see below)
        :param model_id: dynamind model id
        :type filename: str
        :type model_id: int

        .. code-block::

            {
               name: "some name",
               description: "some desc",

               //optionally add assessment model stage of development
               //1 = ALPHA
               //2 = BETA
               //3 = UNDER DEVELOPMENT
               //default is 3
               stage: 2
               //must specify one of:
               model_id: <model_id> //by default will use the active version of this model
               model_version_id: <model_version_id> //if present will use this model version id
            }

        returns:

        .. code-block::

            {
              assessment_model_id: <the id of the new assessment model>,
              assessment_model_version_id: <id of the new default version>
            }

        """

        with open(filename) as json_file:
            node_data = json.load(json_file)

        if model_id is not None:
            node_data["model_id"] = model_id

        r = self._post(self.api_url + "/assessment_models", node_data)

        if r.status_code == 200:
            result = r.json()
            return result["assessment_model_id"]
        raise Exception(f"Unable to create assessment model {r.status_code}, {r.json()}")



    def update_assessment_model(self, assessment_model_id, filename, model_id):
        """
        Creates a new assessment model and a default version tagged as 0.0.1
        the data must be of the shape:

        :param assessment_model_id: assessment model id to be updated
        :param filename: filename of json file (see below)
        :param model_id: dynamind model id
        :type assessment_model_id: int
        :type filename: str
        :type model_id: int

        .. code-block::

            {
               name: "some name",
               description: "some desc",

               //optionally add assessment model stage of development
               //1 = ALPHA
               //2 = BETA
               //3 = UNDER DEVELOPMENT
               //default is 3
               stage: 2
               //must specify one of:
               model_id: <model_id> //by default will use the active version of this model
               model_version_id: <model_version_id> //if present will use this model version id
            }

        returns:

        .. code-block::

            {
              assessment_model_id: <the id of the new assessment model>,
              assessment_model_version_id: <id of the new default version>
            }

        """
        with open(filename) as json_file:
            node_data = json.load(json_file)

        if model_id is not None:
            node_data["model_id"] = model_id

        r = self._post(f"{self.api_url}/assessment_models/{assessment_model_id}/versions", node_data)
        if r.status_code == 200:
            result = r.json()
            return result["assessment_model_version_id"]
        raise Exception(f"Unable to update assessment model {r.status_code}, {r.json()}")

    def get_project_databases(self, project_id, folder=".", scenario_id=None):
        """
        Download project databases. Databases will be downloaded into folder/project_id.zip
        For larger projects it is recommended to defined the scenario_id to be downloaded. Otherwise the download might fail

        :param project_id: project id
        :param folder: folder
        :param scenario_id: scenario_id
        :type project_id: int
        :type folder: str
        :type scenario_id: int
        """

        if scenario_id:
            r = self._get(f"{self.api_url}/projects/{project_id}/data?scenario={scenario_id}")
        else:
            r = self._get(f"{self.api_url}/projects/{project_id}/data")
        if r.status_code == 200:
            if scenario_id:
                open(f"{folder}/{project_id}-{scenario_id}.zip", 'wb').write(r.content)
            else:
                open(f"{folder}/{project_id}.zip", 'wb').write(r.content)
            return
        raise Exception(f"Something went wrong while downloading the folder {r.status_code} {r.json()}")

    def get_my_status(self):
        """
        Get user status

        :return: dict with project status
        """

        r = self._get(f"{self.api_url}/user/status/")
        if r.status_code == 200:
            return r.json()

        raise Exception(f"Something when downloading status {r.status_code} {r.json()}")

    def get_default_parameter_dict(self, node_id: int) -> dict:
        """
        Returns default parameter dict for a node

        :param node_id:
        :type node_id: int
        :return: Parameter dict
        :rtype dict
        """
        nodes = self.get_nodes()
        parameter_dict = {}
        for n in nodes:
            if node_id != n['id']:
                continue
            versions = [(idx, v['id']) for idx, v in enumerate(n['versions'])]
            latest_version = max(versions, key=lambda t: t[1])
            version = n['versions'][latest_version[0]]
            for m in version["models"]:
                for p in m['parameter_description']:
                    if 'fields' in p:
                        for f in p['fields']:
                            parameter_dict[f['parameter']] = f['default']
                    if 'parameter' in p:
                        parameter_dict[p['parameter']] = p['default']
            return parameter_dict
        raise Exception(f"Node not found")

    def show_assessment_models(self) -> None:
        """
        List available assessment models

        :return: None
        """
        models = self.get_assessment_models()
        for m in models['assessment_models']:
            print(m['name'])




from scenario_tool_interface import ScenarioToolInterface

sti = ScenarioToolInterface(api_url="https://development-api.dance4water.org/api" )

sti.login("christian.urich@gmail.com", "rejudo01")

project_id = 30327
scenarios = sti.get_scenarios(project_id)

scenario_id = 0
for s in scenarios:
    scenario_id = s['id']
print()
sti.show_log(scenario_id)


from lxml import etree as ET

import uuid
import json

class Json2DynaMindXML:

    def __init__(self):
        pass

    def _write_gui_node(self, gui_nodes, uuid, pos_x, pos_y = 0 ):
        node = ET.SubElement(gui_nodes, 'GUI_Node')

        ET.SubElement(node, 'GUI_UUID').attrib["value"] = uuid
        ET.SubElement(node, 'GUI_PosX').attrib["value"] = str(pos_x * 300)
        ET.SubElement(node, 'GUI_PosY').attrib["value"] = str(pos_y)
        ET.SubElement(node, 'GUI_Mini').attrib["value"] = str(0)

    def _write_node(self, nodes, group_uuid, n_dict=None):

        node = ET.SubElement(nodes, 'Node')


        class_name = ET.SubElement(node, 'ClassName')
        class_name.attrib["value"] = n_dict['class_name']


        uuid = ET.SubElement(node, 'UUID')
        uuid.attrib["value"] = n_dict['uuid']

        name = ET.SubElement(node, 'Name')
        name.attrib["value"] = n_dict['name']

        group = ET.SubElement(node, 'GroupUUID')
        group.attrib["value"] = group_uuid


        if "parameters" in n_dict:
            for p, val in n_dict["parameters"].items():
                param = ET.SubElement(node, 'parameter')
                param.attrib["name"] = p
                param.text = val

        if "filter" in n_dict:

            filter_view = "".join(n_dict["filter"]["view_name"])
            attribute_filter = "".join(n_dict["filter"]["attribute_filter"])
            spatial_filter = "".join(n_dict["filter"]["spatial_filter"])

            filter = ET.SubElement(node, 'Filter')
            view_name = ET.SubElement(filter, 'view_name')
            view_name.text = filter_view

            af = ET.SubElement(filter, 'attribtue_filter')
            af.text = attribute_filter

            sf = ET.SubElement(filter, 'spatial_filter')
            sf.text = spatial_filter


    def _write_link(self, links, node_from, node_to):

        link = ET.SubElement(links, 'Link')

        ET.SubElement(link, 'BackLink').attrib["value"] = "0"

        inport = ET.SubElement(link, 'InPort')
        ET.SubElement(inport, 'UUID').attrib["value"] = node_to
        ET.SubElement(inport, 'PortName').attrib["value"] = "city"
        ET.SubElement(inport, 'PortType').attrib["value"] = "0"

        outport = ET.SubElement(link, 'OutPort')
        ET.SubElement(outport, 'UUID').attrib["value"] = node_from
        ET.SubElement(outport, 'PortName').attrib["value"] = "city"
        ET.SubElement(outport, 'PortType').attrib["value"] = "0"


    def _write_headers(self, dynamind):

        info = ET.SubElement(dynamind, 'Info')
        info.attrib["Version"] = "0.5"

        dynamind_core = ET.SubElement(dynamind, 'DynaMindCore')
        dynamind_gui = ET.SubElement(dynamind, 'DynaMindGUI')
        dynamind_gui_nodes = ET.SubElement(dynamind_gui, 'GUI_Nodes')

        settings = ET.SubElement(dynamind_core, 'Settings')

        epsg = ET.SubElement(settings, 'EPSG')
        epsg.attrib["value"] = '3857'

        workdir = ET.SubElement(settings, 'WorkingDir')
        workdir.attrib["value"] = '/tmp'

        keep_sys = ET.SubElement(settings, 'KeepSystems')
        keep_sys.attrib["value"] = '0'

        nodes = ET.SubElement(dynamind_core, 'Nodes')

        node = ET.SubElement(nodes, 'RootNode')
        uuid = ET.SubElement(node, 'UUID')
        uuid.attrib["value"] = '0'


        links = ET.SubElement(dynamind_core, 'Links')

        return nodes, links, dynamind_gui_nodes

    def _write_to_file(self, root, file_name):
        with open(file_name, 'w')as dynamind_file:
            xml_str = ET.tostring(root, pretty_print=True).decode('utf-8')
            dynamind_file.write(xml_str)

    def _create_group_nodes(self, nodes, links, gui_nodes, group_uuid, group):

        n_prev = None
        elements_pos = 0
        for n in group["nodes"]:
            next_uuid = str(uuid.uuid4())
            if not n_prev and group_uuid != "0":
                n_prev = group_uuid
            if "group" in n:
                self._write_node(nodes, group_uuid, self._create_default_group(n, next_uuid))
                self._create_group_nodes(nodes, links, gui_nodes, next_uuid, n)

            else:
                n["uuid"] = next_uuid
                self._write_node(nodes, group_uuid, n)
            if (n_prev):
                self._write_link(links, n_prev, next_uuid)
            self._write_gui_node(gui_nodes, next_uuid, elements_pos)

            elements_pos+=1
            n_prev = next_uuid

        # Last Link
        if (n_prev and group_uuid != "0"):
            self._write_link(links, next_uuid, group_uuid)

    def _create_default_group(self, n, next_uuid):
        g = {}
        g["uuid"] = next_uuid
        g["class_name"] = "DMLoopGroup"
        g["name"] = n["group"]
        s = {}
        s["writeStreams"] = "city*|*"
        s["readStreams"] = ""
        g["parameters"] = s
        return g

    def dump(self, input_file, output_file):
        root = ET.Element('DynaMind')
        with open(input_file) as json_file:
            data_structure = json.load(json_file)
        nodes, links, dynamind_gui_nodes = self._write_headers(root)
        self._create_group_nodes(nodes, links, dynamind_gui_nodes, "0", data_structure[0])
        self._write_to_file(root, output_file)















import xml.etree.ElementTree as ET
import json
import copy

class DynaMindXML2Json:

    def __init__(self):
        self._nodes = {}

    def _init_data_structure(self, file_name):
        tree = ET.parse(file_name)
        root = tree.getroot()

        dm_nodes = root.find("DynaMindCore/Nodes")
        dm_links = root.find("DynaMindCore/Links")

        links = {}
        for l in dm_links:
            if l.find('OutPort').find('UUID').get("value") in links:
                vec = links[l.find('OutPort').find('UUID').get("value")]
                vec.append(l.find('InPort').find('UUID').get("value"))
                links[l.find('OutPort').find('UUID').get("value")] = vec
                continue
            links[l.find('OutPort').find('UUID').get("value")] = [l.find('InPort').find('UUID').get("value")]

        self._nodes = {}
        group_uuids = set()

        for n in dm_nodes:
            node = {}
            if n.find('UUID').get("value") == "0":
                continue
            node["name"] = n.find('Name').get("value")
            node["class_name"] = n.find('ClassName').get("value")
            node["uuid"] = n.find('UUID').get("value")
            node["group_uuid"] = n.find('GroupUUID').get("value")

            filter = n.find('Filter')
            if filter:
                filter_view = [e.text.strip() for e in filter.findall("view_name")]
                attribtue_filter = [e.text.strip() if e.text else "" for e in filter.findall("attribtue_filter")]
                spatial_filter = [e.text.strip() if e.text else "" for e in filter.findall("spatial_filter")]
                node_filter = {}
                node_filter["view_name"] = filter_view
                node_filter["attribute_filter"] = attribtue_filter
                node_filter["spatial_filter"] = spatial_filter
                node["filter"] = node_filter
                # print(node["name"], node_filter)
            group_uuids.add(node["group_uuid"])
            node["parameters"] = self._generate_parameters(n)
            self._nodes[n.find('UUID').get("value")] = node

        for uuid_in in links.keys():
            node =self._nodes[uuid_in]
            node["link_to"] = links[uuid_in]

    def _generate_parameters(self, node):
        parameters = {}
        ps = node.findall('parameter')
        for p in ps:
            parameters[p.get("name")] = p.text.strip() if p.text else ""
        return parameters

    def _start_in_group(self, uuid):
        group_nodes = []
        linked_nodes = []
        start_nodes = []
        for k, n in self._nodes.items():
            if n["group_uuid"] != uuid:
                continue
            group_nodes.append(k)
            if "link_to" not in n:
                continue
            for l in n["link_to"]: # Check if it links to same group level
                if self._nodes[l]["group_uuid"] != uuid:
                    continue
                linked_nodes.append(l)
        for id in group_nodes:
            if id not in linked_nodes:
                start_nodes.append(id)

        return start_nodes

    def _is_group(self, uuid):
        if self._nodes[uuid]["class_name"] == "DMLoopGroup":
            return True
        return False

    def _append_node(self, t, node):
        n = copy.deepcopy(node)
        keys = ['uuid', 'group_uuid', 'link_to']
        for k in keys:
            if k in n: del n[k]
        t.append(n)

        if self._is_group(node["uuid"]):
            branch = []
            keys = ['class_name', 'parameters', 'name']

            for k in keys:
                if k in n: del n[k]
            n["group"] = node["name"]
            n["nodes"] = branch
            return branch
        return t

    def _branch(self, tree, node, group_uuid, down=False):
        if not down:
            tree = self._append_node(tree, node)
        if "link_to" not in node:
            return

        for l in node["link_to"]:
            down = False
            if self._nodes[l]["group_uuid"] != group_uuid:
                continue
            if self._is_group(l):
                self._branch(tree, self._nodes[l], l)
                down = True
            self._branch(tree, self._nodes[l], group_uuid, down)

    def dump(self, file_name, output):
        self._init_data_structure(file_name)
        tree = []
        start_nodes = self._start_in_group("0")
        root_node = {}

        for s in start_nodes:
            # append_clean_node(t, nodes[s])
            root_node["name"] = "root"
            root_node["class_name"] = "DMLoopGroup"
            root_node["uuid"] = ""
            root_node["group_uuid"] = "root"
            root_node["parameters"] = ""
            root_node["uuid"] = "0"
            root_node["link_to"] = [s]
            self._nodes["0"] = root_node

            self._branch(tree, root_node, "0")

        with open(output, 'w') as fp:
            json.dump(tree, fp, indent=2, sort_keys=False)
